;----------------------------------------------------------------------------
; File name:	Start_FS.S			Revision date:	1998.02.13
; Created by:	Ulf Ronald Andersson		Creation date:	1998.01.25
;----------------------------------------------------------------------------
; Purpose:
;	Startup file suitable for compiling MetaDos '.DOS' filesystem
;	drivers.  It contains wrapping functions that call corresponding
;	high level C functions for the actual filesystem access.
;	This file may also be used as a skeleton for pure assembly
;	filesystems by editing the wrapping functions so as to insert
;	direct code instead of the wrapped C functions.
;----------------------------------------------------------------------------
; Credits:
;	Thanks go to Julian F. Reschke for producing earlier 'skeletons'
;	which enabled me to learn the principles involved.  Thanks !
;----------------------------------------------------------------------------
	include	"uran\STRUCT.SH"
	include	"uran\URAn_SYS.SH"
	include	"uran\URAn_DOS.SH"
;----------------------------------------------------------------------------
MAX_Func	equ	349
;----------------------------------------------------------------------------
	.globl	hl_bootup
	.globl	hl_initfun,hl_Dfree,hl_Dcreate,hl_Ddelete,hl_Dsetpath
	.globl	hl_Fcreate,hl_Fopen,hl_Fclose,hl_Fread,hl_Fwrite
	.globl	hl_Fdelete,hl_Fseek,hl_Fattrib,hl_Fsfirst,hl_Fsnext
	.globl	hl_Frename,hl_Fdatime
;
	.globl	hl_Fcntl,hl_Dpathconf
	.globl	hl_Dopendir,hl_Dreaddir,hl_Drewinddir,hl_Dclosedir
	.globl	hl_Fxattr,hl_Flink,hl_Fsymlink,hl_Freadlink
	.globl	hl_Dcntl,hl_Dxreaddir,hl_Dreadlabel,hl_Dwritelabel
;----------------------------------------------------------------------------
	.globl	ll_illegal,ll_illegal_flag,ll_Bconws
;----------------------------------------------------------------------------
	.text
;----------------------------------------------------------------------------
ll_bootup:
	bsr	Init_FunctionTable
	bsr		hl_bootup	;boot high level code, returning d0
	move.l	d0,d1			;for the driver name string pointer
	move.l	#FunctionTable+12,d0
	rts
;----------------------------------------------------------------------------
ll_initfun:
	movem.l	a0-a6/d2-d7,-(sp)
	move.l	a2,-(sp)		;push -> DOS parameters
	move.b	d2,-(sp)		;push -> DOS drive char
	move.b	d0,-(sp)		;push -> BOS drive char
	bsr		hl_initfun
	addq	#8,sp
	move.l	d0,a0			;a0 = d0 = ldp -> logical device block
	move.l	(a0),d1			;d1 = device filesystem mode bits
; d1: filename conversion
; Bit 0: 0 = force upper case
; Bit 1: 0 = fully expand
; Bit 2: 0 = critical error handler reports EWRPRO
	movem.l	(sp)+,a0-a6/d2-d7
	rts
;----------------------------------------------------------------------------
ll_INVFN:
	moveq	#E~INVFN,d0
	rts
;----------------------------------------------------------------------------
;Start of:	ll_functions	low level functions wrapping high level ones
;----------------------------------------------------------------------------
ll_Dfree:
	move.l	a5,-(sp)
	move.l	a4,-(sp)
	move.l	a3,-(sp)
	bsr		hl_Dfree
	lea	12(sp),sp
	rts
;----------------------------------------------------------------------------
ll_Dcreate:
	move.l	a5,-(sp)
	move.l	a4,-(sp)
	move.l	a3,-(sp)
	bsr		hl_Dcreate
	lea	12(sp),sp
	rts
;----------------------------------------------------------------------------
ll_Ddelete:
	move.l	a5,-(sp)
	move.l	a4,-(sp)
	move.l	a3,-(sp)
	bsr		hl_Ddelete
	lea	12(sp),sp
	rts
;----------------------------------------------------------------------------
ll_Dsetpath:
	move.l	a5,-(sp)
	move.l	a4,-(sp)
	move.l	a3,-(sp)
	bsr		hl_Dsetpath
	lea	12(sp),sp
	rts
;----------------------------------------------------------------------------
ll_Fcreate:
	move.l	a5,-(sp)
	move.l	a4,-(sp)
	move.l	a3,-(sp)
	bsr		hl_Fcreate
	lea	12(sp),sp
	rts
;----------------------------------------------------------------------------
ll_Fopen:
	move.l	a5,-(sp)
	move.l	a4,-(sp)
	move.l	a3,-(sp)
	bsr		hl_Fopen
	lea	12(sp),sp
	rts
;----------------------------------------------------------------------------
ll_Fclose:
	move.l	a5,-(sp)
	move.l	a4,-(sp)
	move.l	a3,-(sp)
	bsr		hl_Fclose
	lea	12(sp),sp
	rts
;----------------------------------------------------------------------------
ll_Fread:
	move.l	a5,-(sp)
	move.l	a4,-(sp)
	move.l	a3,-(sp)
	bsr		hl_Fread
	lea	12(sp),sp
	rts
;----------------------------------------------------------------------------
ll_Fwrite:
	move.l	a5,-(sp)
	move.l	a4,-(sp)
	move.l	a3,-(sp)
	bsr		hl_Fwrite
	lea	12(sp),sp
	rts
;----------------------------------------------------------------------------
ll_Fdelete:
	move.l	a5,-(sp)
	move.l	a4,-(sp)
	move.l	a3,-(sp)
	bsr		hl_Fdelete
	lea	12(sp),sp
	rts
;----------------------------------------------------------------------------
ll_Fseek:
	move.l	a5,-(sp)
	move.l	a4,-(sp)
	move.l	a3,-(sp)
	bsr		hl_Fseek
	lea	12(sp),sp
	rts
;----------------------------------------------------------------------------
ll_Fattrib:
	move.l	a5,-(sp)
	move.l	a4,-(sp)
	move.l	a3,-(sp)
	bsr		hl_Fattrib
	lea	12(sp),sp
	rts
;----------------------------------------------------------------------------
ll_Fsfirst:
	move.l	a5,-(sp)
	move.l	a4,-(sp)
	move.l	a3,-(sp)
	bsr		hl_Fsfirst
	lea	12(sp),sp
	rts
;----------------------------------------------------------------------------
ll_Fsnext:
	move.l	a5,-(sp)
	move.l	a4,-(sp)
	move.l	a3,-(sp)
	bsr		hl_Fsnext
	lea	12(sp),sp
	rts
;----------------------------------------------------------------------------
ll_Frename:
	move.l	a5,-(sp)
	move.l	a4,-(sp)
	move.l	a3,-(sp)
	bsr		hl_Frename
	lea	12(sp),sp
	rts
;----------------------------------------------------------------------------
ll_Fdatime:
	move.l	a5,-(sp)
	move.l	a4,-(sp)
	move.l	a3,-(sp)
	bsr		hl_Fdatime
	lea	12(sp),sp
	rts
;----------------------------------------------------------------------------
ll_Fcntl:
	move.l	a5,-(sp)
	move.l	a4,-(sp)
	move.l	a3,-(sp)
	bsr		hl_Fcntl
	lea	12(sp),sp
	rts
;----------------------------------------------------------------------------
ll_Dpathconf:
	move.l	a5,-(sp)
	move.l	a4,-(sp)
	move.l	a3,-(sp)
	bsr		hl_Dpathconf
	lea	12(sp),sp
	rts
;----------------------------------------------------------------------------
ll_Dopendir:
	move.l	a5,-(sp)
	move.l	a4,-(sp)
	move.l	a3,-(sp)
	bsr		hl_Dopendir
	lea	12(sp),sp
	rts
;----------------------------------------------------------------------------
ll_Dxreaddir:
	move.l	a5,-(sp)
	move.l	a4,-(sp)
	move.l	a3,-(sp)
	bsr		hl_Dxreaddir
	lea	12(sp),sp
	rts
;----------------------------------------------------------------------------
ll_Dreaddir:
	move.l	a5,-(sp)
	move.l	a4,-(sp)
	move.l	a3,-(sp)
	bsr		hl_Dreaddir
	lea	12(sp),sp
	rts
;----------------------------------------------------------------------------
ll_Drewinddir:
	move.l	a5,-(sp)
	move.l	a4,-(sp)
	move.l	a3,-(sp)
	bsr		hl_Drewinddir
	lea	12(sp),sp
	rts
;----------------------------------------------------------------------------
ll_Dclosedir:
	move.l	a5,-(sp)
	move.l	a4,-(sp)
	move.l	a3,-(sp)
	bsr		hl_Dclosedir
	lea	12(sp),sp
	rts
;----------------------------------------------------------------------------
ll_Fxattr:
	move.l	a5,-(sp)
	move.l	a4,-(sp)
	move.l	a3,-(sp)
	bsr		hl_Fxattr
	lea	12(sp),sp
	rts
;----------------------------------------------------------------------------
ll_Flink:
	move.l	a5,-(sp)
	move.l	a4,-(sp)
	move.l	a3,-(sp)
	bsr		hl_Flink
	lea	12(sp),sp
	rts
;----------------------------------------------------------------------------
ll_Fsymlink:
	move.l	a5,-(sp)
	move.l	a4,-(sp)
	move.l	a3,-(sp)
	bsr		hl_Fsymlink
	lea	12(sp),sp
	rts
;----------------------------------------------------------------------------
ll_Freadlink:
	move.l	a5,-(sp)
	move.l	a4,-(sp)
	move.l	a3,-(sp)
	bsr		hl_Freadlink
	lea	12(sp),sp
	rts
;----------------------------------------------------------------------------
ll_Dcntl:
	move.l	a5,-(sp)
	move.l	a4,-(sp)
	move.l	a3,-(sp)
	bsr		hl_Dcntl
	lea	12(sp),sp
	rts
;----------------------------------------------------------------------------
ll_Dreadlabel:
	move.l	a5,-(sp)
	move.l	a4,-(sp)
	move.l	a3,-(sp)
	bsr		hl_Dreadlabel
	lea	12(sp),sp
	rts
;----------------------------------------------------------------------------
ll_Dwritelabel:
	move.l	a5,-(sp)
	move.l	a4,-(sp)
	move.l	a3,-(sp)
	bsr		hl_Dwritelabel
	lea	12(sp),sp
	rts
;----------------------------------------------------------------------------
;End of:	ll_functions	low level functions wrapping high level ones
;----------------------------------------------------------------------------
;Start of:	General Subroutines
;----------------------------------------------------------------------------
Init_FunctionTable:
	lea	InitTable(pc),a0
	lea	FunctionTable(pc),a1
	move.l	a1,a2
	move	#(MAX_Func+5)-1,d0
	move.l	#ll_INVFN,d1
.loop_1:				;loop back for primary init
	move.l	d1,(a2)+		;mark function as invalid
	dbra	d0,.loop_1		;loop back for primary init completion
;
	move	(a0)+,d0		;d0 = longs of first block
.loop_2:				;loop start for each block
	subq	#1,d0			;d0 = adapted for dbra
	move	(a0)+,d1		;d1 = index relative to function 0
	addq	#4,d1			;d1 = index relative to table base
	asl	#2,d1			;d1 = offset relative to table base
	lea	(a1,d1.w),a2		;a2 -> current destination block
.loop_3:				;loop start for block copy
	move.l	(a0)+,(a2)+		;copy one long of block
	dbra	d0,.loop_3		;loop back for entire block
	move	(a0)+,d0		;d0 = longs of next block
	bne.s	.loop_2			;loop back unless empty block (end mark)
	rts
;----------------------------------------------------------------------------
;static int ll_Bconws (char *str)
;
ll_Bconws:
	movem.l	a2-a3,-(sp)
	move.l	a0,a3
	move.l	a0,d0
	beq.s	.exit
	bra.s	.loop_entry
;
.loop:
	bios	Bconout,#2,#CR
.loop_entry:
	move.b	(a3)+,d0
	bne.s	.loop
	subq	#1,a3
.exit:
	move.l	a3,d0
	sub.l	a0,d0
	movem.l	(sp)+,a2-a3
	rts
;----------------------------------------------------------------------------
ll_illegal:
ll_illegal_flag:
	illegal
	rts
;----------------------------------------------------------------------------
;End of:	General Subroutines
;----------------------------------------------------------------------------
	.data
;----------------------------------------------------------------------------
InitTable:
	dc.w	4,-4
	dc.l		'MAGI','CMET',MAX_Func,ll_initfun
	dc.w	1,(Dfree & $ff)
	dc.l		ll_Dfree
	dc.w	3,(Dcreate & $ff)
	dc.l		ll_Dcreate,ll_Ddelete,ll_Dsetpath
	dc.w	8,(Fcreate & $ff)
	dc.l		ll_Fcreate,ll_Fopen,ll_Fclose,ll_Fread
	dc.l		ll_Fwrite,ll_Fdelete,ll_Fseek,ll_Fattrib
	dc.w	2,(Fsfirst & $ff)
	dc.l		ll_Fsfirst,ll_Fsnext
	dc.w	1,(Frename & $ff)
	dc.l		ll_Frename
	dc.w	1,(Fdatime & $ff)
	dc.l		ll_Fdatime
	dc.w	1,260
	dc.l		ll_Fcntl
	dc.w	1,292
	dc.l		ll_Dpathconf
	dc.w	9,296
	dc.l		ll_Dopendir,ll_Dreaddir,ll_Drewinddir,ll_Dclosedir
	dc.l		ll_Fxattr,ll_Flink,ll_Fsymlink,ll_Freadlink
	dc.l		ll_Dcntl
	dc.w	1,322
	dc.l		ll_Dxreaddir
	dc.w	1,338
	dc.l		ll_Dreadlabel
	dc.w	1,339
	dc.l		ll_Dwritelabel
	dc.w	0	
;----------------------------------------------------------------------------
	.bss
;----------------------------------------------------------------------------
FunctionTable:
	ds.l	MAX_Func+5
;----------------------------------------------------------------------------
	.end
;----------------------------------------------------------------------------
; End of file:	Start_FS.S
;----------------------------------------------------------------------------
