/*--------------------------------------------------------------------------*/
/* File name:		ECHO_SRV.C					Revision date:	1999.04.13	*/
/* Creator:			Ulf Ronald Andersson		Creation date:	1998.11.27	*/
/* (c)1996 by:		Ulf Ronald Andersson		All rights reserved			*/
/* Feedback to:		dlanor@oden.se											*/
/*--------------------------------------------------------------------------*/
#include	<stdio.h>
#include	<tos.h>
#include	<string.h>
#include	<time.h>
#include	<ctype.h>
#include	<stdlib.h>
#include	<screen.h>
/*----*/
#include	<sting\transprt.h>
#include	<sting\port.h>
#include	<sting\layer.h>
#include	<sting\netd.h>
/*--------------------------------------------------------------------------*/
#define	M_TITLE		"ECHO server"
#define	M_VERSION	"01.04"
#define	M_YEAR		1999
#define	M_MONTH		04
#define	M_DAY		13
#define	M_AUTHOR	"Ronald Andersson"
/*--------------------------------------------------------------------------*/
#define	ECHO_PORT	7
#define	NetD_UDP	0			/* UDP ports will use type_bsize 0		*/
#define	NetD_TCP	1024		/* TCP ports will use type_bsize 1024	*/
/*--------------------------------------------------------------------------*/
DRV_LIST	*sting_drivers;
TPL			*tpl;
STX			*stx;
NetD_STX	*NetD_STX_p;
NetD_SRV	*my_SRV;
/*--------------------------------------------------------------------------*/
/* Start of:		Subfunctions for initialization							*/
/*--------------------------------------------------------------------------*/
long  get_sting_cookie(void)
{	long	*work;
	for (work = * (long **) 0x5a0L; *work != 0L; work += 2)
		if (*work == 'STiK')
			return (*++work);
	return (0L);
}
/*--------------------------------------------------------------------------*/
long  get_netd_cookie(void)
{	long	*work;
	for (work = * (long **) 0x5a0L; *work != 0L; work += 2)
		if (*work == 'NetD')
			return (*++work);
	return (0L);
}
/*--------------------------------------------------------------------------*/
int 	is_STING (void)
{	int 	res = 0;

	Cur_home();
	Next_line();	/* Do not overwrite the menu bar */

	sting_drivers = (DRV_LIST *) Supexec (get_sting_cookie);

	if (sting_drivers == 0L)
		return (res);

	if (strcmp (sting_drivers->magic, MAGIC) != 0)
		return (res);

	stx = (STX *) (*sting_drivers->get_dftab) (MODULE_DRIVER);
	if (stx == (STX *) NULL)
		return (res);

	tpl = (TPL *) (*sting_drivers->get_dftab) (TRANSPORT_DRIVER);
	if (tpl == (TPL *) NULL)
		return (res);

	return (TRUE);
}
/*--------------------------------------------------------------------------*/
int 	is_NETD (void)
{	int 	res = 0;

	NetD_STX_p = (NetD_STX *) Supexec (get_netd_cookie);

	if	(NetD_STX_p == (NetD_STX *) NULL)
		return (res);

	return (TRUE);
}
/*--------------------------------------------------------------------------*/
void	wmess (const char *s)	/* Write string to the desktop & keep cnt	*/
{	static	int	lines = 0;

	Cconws (s);
	Next_line();

	if (++lines > 36)
	{	Cur_home();
		Next_line();
		lines = 0;
	}
}
/*--------------------------------------------------------------------------*/
/* End of:			Subfunctions for initialization							*/
/*--------------------------------------------------------------------------*/
/* Start of:		server functions called by NetD STX						*/
/*--------------------------------------------------------------------------*/
int16	SRV_connect (NetD_APP *NetD_APP_p, NetD_CON *new_conn_p)
{	/* This server does no work on connection, so function is almost empty	*/
	(void)	NetD_APP_p;		/* We don't use that argument */

	((int32 *)(new_conn_p->pad))[0] = TIMER_now();
	return (E_NORMAL);	/* always accepts connection */
}	/* End of: SRV_connect */
/*--------------------------------------------------------------------------*/
void	SRV_traffic (NetD_APP *NetD_APP_p)
{	NetD_CON	*work, *next;
	NDB			*ndb;
	(void)	NetD_APP_p;		/* We don't use that argument */

	for	(work = my_SRV->active_q; work; work = next)
	{	while	((ndb = CNget_NDB(work->handle)) > 0)
		{	NetD_send_CON(work, ndb->ndata, ndb->len);	/* echo the data */
			KRfree(ndb->ptr);
			KRfree(ndb);
			((int32 *)(work->pad))[0] = TIMER_now();
		}
		next = work->next;
		if	( TIMER_elapsed(((int32 *)(work->pad))[0]) > 10000 )
			NetD_kill_CON( my_SRV, work );	/* close at 10 sec silence */
	}
}	/* End of: SRV_traffic */
/*--------------------------------------------------------------------------*/
void	SRV_warning (NetD_APP *NetD_APP_p)
{	/* This server ignores warning call, so the function is empty */
	(void)	NetD_APP_p;		/* We don't use that argument */

}	/* End of: SRV_warning */
/*--------------------------------------------------------------------------*/
/* End of:			server functions called by NetD STX						*/
/*--------------------------------------------------------------------------*/
/* Start of:		main program function, used only for server init		*/
/*--------------------------------------------------------------------------*/
int 	main (void)
{	NetD_CON	*test_CON_p;

	if	(!is_STING())
	{	wmess("STiNG not installed!! | Aborting Installation...");
		return (0);
	}
	if	(!is_NETD())
	{	wmess("NetD not installed!! | Aborting Installation...");
		return (0);
	}

	my_SRV = (NetD_SRV *) _BasPag;

	my_SRV->if_ver	= REF_IF_VER;
	my_SRV->date	= ((M_YEAR-1980)<<9)|(M_MONTH<<5)|M_DAY;
	my_SRV->version	= M_VERSION;
	my_SRV->author	= M_AUTHOR;
	my_SRV->name	= M_TITLE;

	if	(NetD_init_SRV(my_SRV) < 0)
	{	wmess("NetD refused server! | Aborting installation...");
		return (0);
	}

	/* The stuff below is specific for each server */

	test_CON_p = NetD_init_CON(my_SRV, NetD_UDP, ECHO_PORT);	/* UDP port	*/
	if	((void *) test_CON_p == NULL)
		wmess("NetD refused using UDP ECHO_PORT !");
	test_CON_p = NetD_init_CON(my_SRV, NetD_TCP, ECHO_PORT);	/* TCP port	*/
	if	((void *) test_CON_p == NULL)
		wmess("NetD refused using TCP ECHO_PORT !");

	Ptermres (_PgmSize, 0);		/* we go resident here, so... */
	return (0);					/* we never come here... */
}
/*--------------------------------------------------------------------------*/
/* End of:			main program function, used only for server init		*/
/*--------------------------------------------------------------------------*/
/* End of file:		ECHO_SRV.C												*/
/*--------------------------------------------------------------------------*/
