; /-----------------------------------------------------------------------/
; file:		tcpip.h
; function:	Transport Layer Library defined
; opcode:		none
; part of:	STiNG
;
; author:		E Kuik
; company:	Merciful Software (1997)
; /-----------------------------------------------------------------------/

; Ronald:
; This is the TCP/IP lib including I'm using.
; You may need it for the macro's.

/*
 *   Driver access structure / functions
 */
;drv_header	->	Header part of TPL structure
module	=	0	; Specific string that can be searched for
author	=	4	; Any string
version	=	8	; Format `00.00' Version:Revision


; drv_list	->	This pointer is returned by the cookie value
magic	=	0	; 10 byte Magic string, def'd as "STiKmagic"
get_dftab	=	10	; Get Driver Function Table (pointer)
			; pointer to "TRANSPORT_TCPIP",0 on stack
			; returns pointer to tpl in d0
etm_exec	=	14	; exec a module (pointer) request
			; module name pointer on stack
cfg	=	18	; pointer to Config structure
basepage	=	22	; pointer to STinG basepage (not valid with STiK)

/*
 *   IP packet header.
 */
ipph_version	=	0	; IP Version
ipph_len		=	2	; Internet Header Length
ipph_tos		=	4	; Type of Service
ipph_length	=	6	; Total of all header, options and data
ipph_ident	=	8	; Identification for fragmentation
ipph_reserved	=	10	; Reserved : Must be zero
ipph_dont_frg	=	12	; Don't fragment flag
ipph_more_frg	=	14	; More fragments flag
ipph_frg_ofst	=	16	; Fragment offset
ipph_ttl		=	18	; Time to live
ipph_protocol	=	19	; Protocol
ipph_chksum	=	20	; Header checksum
ipph_ip_src	=	22	; Source IP address
ipph_ip_dest	=	26	; Destination IP address
ipph		=	30	; Header Size (used by ipp calculatons)

/*
 *   Internal IP packet representation.
 */
ipp_options	=	ipph+0	; Options data block (pointer)
ipp_optlength	=	ipph+4	; Length of options data block
ipp_pktdata	=	ipph+6	; IP packet data block (pointer)
ipp_pktlength	=	ipph+10	; Length of IP packet data block (word)
ipp_timeout	=	ipph+12	; Timeout of packet life
ipp_ip_gateway	=	ipph+16	; Gateway for forwarding this packet
ipp_recvd		=	ipph+20	; Receiving port (pointer)
ipp_ippnext	=	ipph+24	; Next IP packet in IP packet queue (pointer)

/*
 *   Input queue structures
 *   Network Data Block.  For data delivery
 */
ndb_ptr		=	0	; Pointer to base of block. (For KRfree();)
ndb_ndata		=	4	; Pointer to next data to deliver
ndb_length	=	8	; Length of remaining data
ndb_next		=	10	; Next NDB in chain or NULL

/*
 *   Connection information block
 */
cib_protocol	=	0	; TCP or UDP or ... 0 means CIB is not in use
cib_lport		=	2	; TCP local  port     (ie: local machine)
cib_rport		=	4	; TCP remote port     (ie: remote machine)
cib_rhost		=	6	; TCP remote IP addr  (ie: remote machine)
cib_lhost		=	10	; TCP local  IP addr  (ie: local machine)
cib_status	=	14	; Net status. 0 means normal

/*
 *   Values for protocol field
 */
P_ICMP	=	1	; IP assigned number for ICMP
P_TCP	=	6	; IP assigned number for TCP
P_UDP	=	17	; IP assigned number for UDP

/*
 *   Transport structure / functions
 */
; the same as in drv_header
;module		=	0	; Specific string that can be searched for
;author		=	4	; Any string
;version		=	8	; Format `00.00' Version:Revision

fKRmalloc		=	12
fKRfree		=	16
fKRgetfree	=	20
fKRrealloc	=	24
fget_err_text	=	28
fgetvstr		=	32
fcarrier_detect	=	36
fTCP_open		=	40
fTCP_close	=	44
fTCP_send		=	48
fTCP_wait_state	=	52
fTCP_ack_wait	=	56
fUDP_open		=	60
fUDP_close	=	64
fUDP_send		=	68
fCNkick		=	72
fCNbyte_count	=	76
fCNget_char	=	80
fCNget_NDB	=	84
fCNget_block	=	88
fhousekeep	=	92
fresolve		=	96
fser_disable	=	100
fser_enable	=	104
fset_flag		=	108
fclear_flag	=	112
fCNgetinfo	=	116
fon_port		=	120
foff_port		=	124
fsetvstr		=	128	; watch out by using setvstr and
				; next! By using: setvstr(a0,d0.w)
				; byte is neg! Use: setvstr(a0)
				; where setvstr is a word
fquery_port	=	132
fCNgets		=	136 
fICMP_send	=	140
fICMP_handler	=	144
fICMP_discard	=	148

/*
 *   Handler flag values.
 */
HNDLR_SET		=	0	; Set new handler if space
HNDLR_FORCE	=	1	; Force new handler to be set
HNDLR_REMOVE	=	2	; Remove handler entry
HNDLR_QUERY	=	3	; Inquire about handler entry

/*
 *   TCP connection states
 */
TCLOSED       =	0    /* No connection.  Null, void, absent, ...      */
TLISTEN       =	1    /* Wait for remote request                      */
TSYN_SENT     =	2    /* Connect request sent, await matching request */
TSYN_RECV     =	3    /* Wait for connection ack                      */
TESTABLISH    =	4    /* Connection established, handshake completed  */
TFIN_WAIT1    =	5    /* Await termination request or ack             */
TFIN_WAIT2    =	6    /* Await termination request                    */
TCLOSE_WAIT   =	7    /* Await termination request from local user    */
TCLOSING      =	8    /* Await termination ack from remote TCP        */
TLAST_ACK     =	9    /* Await ack of terminate request sent          */
TTIME_WAIT   =	10   /* Delay, ensures remote has received term' ack */

/*
 *   Error return values
 */
E_NORMAL        =	 0     /* No error occured ...                     */
E_OBUFFULL      =	-1     /* Output buffer is full                    */
E_NODATA        =	-2     /* No data available                        */
E_EOF           =	-3     /* EOF from remote                          */
E_RRESET        =	-4     /* Reset received from remote               */
E_UA            =	-5     /* Unacceptable packet received, reset      */
E_NOMEM         =	-6     /* Something failed due to lack of memory   */
E_REFUSE        =	-7     /* Connection refused by remote             */
E_BADSYN        =	-8     /* A SYN was received in the window         */
E_BADHANDLE     =	-9     /* Bad connection handle used.              */
E_LISTEN        =	-10    /* The connection is in LISTEN state        */
E_NOCCB         =	-11    /* No free CCB's available                  */
E_NOCONNECTION  =	-12    /* No connection matches this packet (TCP)  */
E_CONNECTFAIL   =	-13    /* Failure to connect to remote port (TCP)  */
E_BADCLOSE      =	-14    /* Invalid TCP_close() requested            */
E_USERTIMEOUT   =	-15    /* A user function timed out                */
E_CNTIMEOUT     =	-16    /* A connection timed out                   */
E_CANTRESOLVE   =	-17    /* Can't resolve the hostname               */
E_BADDNAME      =	-18    /* Domain name or dotted dec. bad format    */
E_LOSTCARRIER   =	-19    /* The modem disconnected                   */
E_NOHOSTNAME    =	-20    /* Hostname does not exist                  */
E_DNSWORKLIMIT  =	-21    /* Resolver Work limit reached              */
E_NONAMESERVER  =	-22    /* No nameservers could be found for query  */
E_DNSBADFORMAT  =	-23    /* Bad format of DS query                   */
E_UNREACHABLE   =	-24    /* Destination unreachable                  */
E_DNSNOADDR     =	-25    /* No address records exist for host        */
E_NOROUTINE     =	-26    /* Routine unavailable                      */
E_LOCKED        =	-27    /* Locked by another application            */
E_FRAGMENT      =	-28    /* Error during fragmentation               */
E_TTLEXCEED     =	-29    /* Time To Live of an IP packet exceeded    */
E_PARAMETER     =	-30    /* Problem with a parameter                 */
E_BIGBUF        =	-31    /* Input buffer is too small for data       */
E_LASTERROR     =	 31    /* ABS of last error code in this list      */

/*
 * FLagbox Flag definitions (max 64 flags ...)
 */
FL_do_resolve	=	0
FL_housekeep	=	1
FL_slipin		=	2
FL_slipout	=	3
FL_tcp_timer	=	4
FL_frag_ttl_check	=	5
FL_g_resv		=	6
FL_g_wait		=	7

/*
 * STIK global configuration structure
 */
cfg_client_ip	=	0	; IP address of client (local) machine
cfg_provider	=	4	; IP address of provider, or 0L
cfg_ttl		=	8	; Default TTL for normal packets
cfg_ping_ttl	=	10	; Default TTL for 'ping'ing
cfg_mtu		=	12	; Default MTU (Maximum Transmission Unit)
cfg_mss		=	14	; Default MSS (Maximum Segment Size)
cfg_df_bufsize	=	16	; Size of defragmentation buffer to use
cfg_rcv_window	=	18	; TCP receive window
cfg_def_rtt	=	20	; Initial RTT time in ms
cfg_time_wait	=	22	; How long to wait in 'TIME_WAIT' state
cfg_unreach_rsp	=	24	; Response to unreachable local ports
cfg_cn_time	=	26	; Time connection was made
cfg_cd_valid	=	30	; Is Modem CD a valid signal ??
cfg_line_pro	=	32	; What type of connection is this (protocol)
cfg_old_vec	=	34	; Old vector address (pointer)
cfg_slip		=	38	; Slip structure for happiness (pointer)
cfg_cfg		=	42	; Space for extra config variables (pointer)
cfg_reports	=	46	; Problem reports printed to screen ?? 
cfg_maxnm_ports	=	48	; Maximum number of ports supported
cfg_received_d	=	50	; Counter for data being received
cfg_sent_d	=	54	; Counter for data being sent


_trnsp_malloc		=	0*4+fKRmalloc
_trnsp_free		=	1*4+fKRmalloc
_trnsp_getfree		=	2*4+fKRmalloc
_trnsp_realloc		=	3*4+fKRmalloc
_trnsp_get_errtext		=	4*4+fKRmalloc
_trnsp_get_vstr		=	5*4+fKRmalloc
_trnsp_carrier_detect	=	6*4+fKRmalloc
_trnsp_TCP_open		=	7*4+fKRmalloc
_trnsp_TCP_close		=	8*4+fKRmalloc
_trnsp_TCP_send		=	9*4+fKRmalloc
_trnsp_TCP_wait_state	=	10*4+fKRmalloc
_trnsp_TCP_ack_wait		=	11*4+fKRmalloc
_trnsp_UDP_open		=	12*4+fKRmalloc
_trnsp_UDP_close		=	13*4+fKRmalloc
_trnsp_UDP_send		=	14*4+fKRmalloc
_trnsp_CN_kick		=	15*4+fKRmalloc
_trnsp_CN_bcount		=	16*4+fKRmalloc
_trnsp_CN_getchar		=	17*4+fKRmalloc
_trnsp_CN_getNDB		=	18*4+fKRmalloc
_trnsp_CN_getblock		=	19*4+fKRmalloc
_trnsp_housekeep		=	20*4+fKRmalloc
_trnsp_resolve		=	21*4+fKRmalloc
_trnsp_ser_disable		=	22*4+fKRmalloc
_trnsp_ser_enable		=	23*4+fKRmalloc
_trnsp_set_flag		=	24*4+fKRmalloc
_trnsp_clr_flag		=	25*4+fKRmalloc
_trnsp_CN_getinfo		=	26*4+fKRmalloc
_trnsp_on_port		=	27*4+fKRmalloc
_trnsp_off_port		=	28*4+fKRmalloc
_trnsp_set_vstr		=	29*4+fKRmalloc
_trnsp_query_port		=	30*4+fKRmalloc


; macro's to access the transport layer
;
; support functions

.macro tcpip_save_reg
	Push	d2-d7/a2-a6
.endm

.macro tcpip_restore_reg
	Pull	d2-d7/a2-a6
.endm

.macro Call_tcp function
	jsr	function
.endm


; access functions

.macro KRmalloc memsize,membase
	tcpip_save_reg
	move.l	memsize,d0
	Call_tcp	_KRmalloc
	tcpip_restore_reg
	move.l	d0,membase
.endm

.macro KRfree memory
	tcpip_save_reg
	move.l	memory,a0
	Call_tcp	_KRfree
	tcpip_restore_reg
.endm

.macro KRgetfree flag,free
	tcpip_save_reg
	move.l	flag,d0
	Call_tcp	_KRgetfree
	tcpip_restore_reg
	move.l	d0,free
.endm

.macro KRrealloc oldmem,newmem
	tcpip_save_reg
	move.l	oldmem,a0
	move.l	newmem,d0
	Call_tcp	_KRrealloc
	tcpip_restore_reg
	move.l	d0,oldmem
.endm

.macro get_err_text code,str
	tcpip_save_reg
	move.w	code,d0
	Call_tcp	_get_err_text
	tcpip_restore_reg
	move.l	d0,str
.endm

.macro getvstr var,value,status
	tcpip_save_reg
	move.l	var,a0
	Call_tcp	_getvstr
	tcpip_restore_reg
	move.l	d0,value
	move.l	d1,status
.endm

.macro carrier_detect carrier
	tcpip_save_reg
	Call_tcp	_carrier_detect
	move.l	d0,carrier
	tcpip_restore_reg
.endm

.macro TCP_open rhost,rport,tos,obsize,handle
	tcpip_save_reg
	move.l	rhost,d0
	move.w	rport,d1
	move.w	tos,d2
	move.w	obsize,d3
	Call_tcp	_TCP_open
	tcpip_restore_reg
	move.w	d0,handle
.endm

.macro TCP_close handle,timeout,status
	tcpip_save_reg
	move.w	handle,d0
	move.w	timeout,d1
	Call_tcp	_TCP_close
	tcpip_restore_reg
	move.w	d0,status
.endm

.macro TCP_send handle,buff,len,status
	tcpip_save_reg
	move.w	handle,d0
	move.l	buff,a0
	move.w	len,d1
	Call_tcp	_TCP_send
	tcpip_restore_reg
	move.w	d0,status
.endm

.macro TCP_wait_state handle,state,timeout,status
	tcpip_save_reg
	move.w	handle,d0
	move.w	state,d1
	move.w	timeout,d2
	Call_tcp	_TCP_state
	tcpip_restore_reg
	move.w	d0,status
.endm

.macro TCP_ack_wait handle,timeout,status
	tcpip_save_reg
	move.w	handle,d0
	move.w	timeout,d1
	Call_tcp	_TCP_ack_wait
	tcpip_restore_reg
	move.w	d0,status
.endm

.macro UDP_open rhost,rport,handle
	tcpip_save_reg
	move.l	rhost,d0
	move.w	rport,d1
	Call_tcp	_UDP_open
	tcpip_restore_reg
	move.w	d0,handle
.endm

.macro UDP_close handle,status
	tcpip_save_reg
	move.w	handle,d0
	Call_tcp	_UDP_close
	tcpip_restore_reg
	move.w	d0,status
.endm

.macro UDP_send handle,buff,len,status
	tcpip_save_reg
	move.w	handle,d0
	move.l	buff,a0
	move.w	len,d1
	Call_tcp	_UDP_send
	tcpip_restore_reg
	move.w	d0,status
.endm

.macro CNkick handle,status
	tcpip_save_reg
	move.w	handle,d0
	Call_tcp	_CNkick
	tcpip_restore_reg
	move.w	d0,status
.endm

.macro CNbyte_count handle,bytes
	tcpip_save_reg
	move.w	handle,d0
	Call_tcp	_CNbyte_count
	tcpip_restore_reg
	move.w	d0,bytes
.endm

.macro CNget_char handle,char
	tcpip_save_reg
	move.w	handle,d0
	Call_tcp	_CNget_char
	tcpip_restore_reg
	move.w	d0,char
.endm

.macro CNget_NDB handle,block
	tcpip_save_reg
	move.w	handle,d0
	Call_tcp	_CNget_NDB
	tcpip_restore_reg
	move.l	d0,block
.endm

.macro CNget_block handle,block,len,status
	tcpip_save_reg
	move.w	handle,d0
	move.l	block,a0
	move.w	len,d1
	Call_tcp	_CNget_block
	tcpip_restore_reg
	move.w	d0,status
.endm

.macro housekeep
	tcpip_save_reg
	Call_tcp	_housekeep
	tcpip_restore_reg
.endm

.macro resolve dn,rdn,alist,lsize,status
	tcpip_save_reg
	move.l	dn,a0
	move.l	rdn,a1
	move.l	alist,a2
	move.w	lsize,d0
	Call_tcp	_resolve
	tcpip_restore_reg
	move.w	d0,status
.endm

.macro ser_disable
	tcpip_save_reg
	Call_tcp	_ser_disable
	tcpip_restore_reg
.endm

.macro ser_enable
	tcpip_save_reg
	Call_tcp	_ser_enable
	tcpip_restore_reg
.endm

.macro set_flag flag,status
	tcpip_save_reg
	move.w	flag,d0
	Call_tcp	_set_flag
	tcpip_restore_reg
	move.w	d0,status
.endm

.macro clear_flag flag,status
	tcpip_save_reg
	move.w	flag,d0
	Call_tcp	_clear_flag
	tcpip_restore_reg
	move.w	#1,status
.endm

.macro CNgetinfo handle,info
	tcpip_save_reg
	move.w	handle,d0
	Call_tcp	_CNgetinfo
	tcpip_restore_reg
	move.l	d0,info
.endm

.macro on_port port,status
	tcpip_save_reg
	move.l	port,a0
	Call_tcp	_on_port
	tcpip_restore_reg
	move.w	d0,status
.endm

.macro off_port port,status
	tcpip_save_reg
	move.l	port,a0
	Call_tcp	_off_port
	tcpip_restore_reg
	move.w	d0,status
.endm

.macro setvstr vs,value,status
	tcpip_save_reg
	move.l	vs,a0
	move.l	value,a1
	Call_tcp	_setvstr
	tcpip_restore_reg
	move.w	d0,status
.endm

.macro query_port port,status
	tcpip_save_reg
	move.l	port,a0
	Call_tcp	_query_port
	tcpip_restore_reg
	move.w	d0,status
.endm
