;----------------------------------------------------------------------------
;File name:	PASS_AVA.S			Revision date:	1997.08.06
;Creator:	Ulf Ronald Andersson		Creation date:	1997.08.01
;(c)1996 by:	Ulf Ronald Andersson		All rights reserved
;----------------------------------------------------------------------------
;Required header declarations:
;
	.include	"uran\STRUCT.SH"	;PASM adapted structures
	.include	"uran\URAn_SYS.SH"	;Main system definitions
	.include	"uran\URAn_LA.SH"	;Line A variables etc
	.include	"uran\URAn_DOS.SH"	;GEMDOS, BIOS, XBIOS
	.include	"uran\URAn_GEM.SH"	;VDI & AES
	.include	"uran\URAn_SIM.SH"	;some AES funcs for TOS/TTP
	.include	"uran\URAn_APP.SH"	;General application support
	.include	"uran\URAn_JAR.SH"	;Cookie jar handling
	.include	"uran\URAn_AV.SH"	;AV/VA protocol + MP protocol
	.include	"uran\URAn_XB.SH"	;XBRA protocol
;
	_uniref		AV_send
	_uniref		VA_send
;
;----------------------------------------------------------------------------
;
MAX_VA_ARGLEN	equ	512
;
;
;Add application dependent assemblytime constants/variables above
;
;----------------------------------------------------------------------------
;
;;;huge_program		= 1	;uncomment if TEXT+DATA sections are > 32KB
;;;keep_windows		= 1	;uncomment to keep ACC vwk in main eventloop
;;;exec_timeout		= 100	;uncomment to enable timer in main eventloop
exec_message		= 1	;uncomment for messages in main eventloop
;
;----------------------------------------------------------------------------
;
	.text
;
	start_app
;
;----------------------------------------------------------------------------
;
	.ifeq	huge_program
	.data
	.endif
;
;Edit the string below, to match the accessory title you want in ACC mode
;
acc_name:
	dc.b	'  Pass AVA',NUL	;for ACC menu registration
	dc.b	NUL
	even
;
;Also edit the string below, to match the application name, without
;file extension, padded to 8 bytes.  Normally this will be overwritten
;by runtime info from 'shel_read' during init, but do it just in case...
;
id_app_name_s:
	dc.b	'PASS_AVA',NUL
	even
;
;----------------------------------------------------------------------------
;
	.ifeq	huge_program
	.bss
	.endif
;
message:	ds.w	8		;evnt_mesag message pipe
intin:		ds.w	30		;30 words or more for AES/VDI
intout:		ds.w	45		;45 words or more for AES/VDI
ptsin:		ds.w	30		;30 words or more for VDI
ptsout:		ds.w	12		;12 words or more for VDI output coordinates
		ds.l	200		;subroutine stack >= 100 longs
mystack:	ds.l	1		;top of subroutine stack
;
;----------------------------------------------------------------------------
;
	.text
;
;----------------------------------------------------------------------------
;	Constants, Variables, and Usage of the URAn_APP.S library
;
;rz		'relative zero' for internal references
;basepage_p	-> program's base page
;tsr_size	.L size of BASEPAGE+TEXT+DATA+BSS  (only TOS/TTP may TSR)
;progtype	.L indicates program type:
;  ...			$0000.$FFFF == TOS/TTP/boot_PRG
;  ...			$0000.ap_id == APP/GTP/non_boot_PRG
;  ...			$FFFF.ap_id == ACC
;acc_flag	is the high word of 'progtype'
;tos_flag	is the low word of 'progtype'
;bootflag	.W $FFFF == PRG booted from AUTO (also sets 'tos_flag')
;menu_id	.W menu id for an ACC, otherwise null
;g_handle	.W workstation handle for ACC/APP/GTP/non_boot_PRG
;vwk_handle	.W virtual workstation handle
;contrl		12 words for AES/VDI control data
;
;wk_x_rez \
;wk_y_rez  \/ from
;wk_pix_w  /\ opnvwk
;wk_pix_h /
;
;MiNT_p		.L NULL or -> MiNT structure
;MagX_p		.L NULL or -> MagX structure
;nAES_p		.L NULL or -> nAES structure
;Gnva_p		.L NULL or -> Gnva structure
;
;line_a_base_p	.L -> line_a variable base
;kbshift_p	.L -> kbshift byte of OS
;currbp_p_p	.L -> OS var -> current basepage
;
;NB: if symbol 'huge_program' is defined, above data is in TEXT section (else in BSS)
;NB: defining 'huge_program' also makes function calls use 'jsr' (instead of bsr)
;
;Required user routines:
;
;init_app	called for ACC/APP/GTP/non_boot_PRG to init application,
;		but doesn't need appl_init, graf_handle, or v_opnvwk,
;		nor does an ACC need menu_register.
;		Suitable use is for initialization of object trees.
;NB:  for ACC	menu_register is called between init_app and exec_app
;
;exec_app	called to execute the main application regardless of type,
;		but doesn't need v_clsvwk, or appl_exit, or ACC appl_mesag.
;		This call will be repeated for a reactivated ACC.
;		Non_acc programs should have exit code in d0 at RTS.
;		(See details at 'Exit codes:' below)
;
;	At entry to either of these two routines:
;
;		d6.W	== bootflag	\
;		d7.L	== progtype	 > See descriptions above.
;		a6.L	== rz		/
;
;Optional user routines:
;
;exec_timer	Called for ACC that has a defined constant 'exec_timeout',
;		whenever evnt_multi produces such a timer event.
;		The constant is placed as long at 'main_timeout', and may
;		there be dynamically adjusted by the program.
;
;exec_mesag	Called for ACC that has a defined constant 'exec_message',
;		whenever evnt_multi/evnt_mesag produces messages that are
;		not AC_OPEN  (such as VA_START).
;
;		If the constant 'keep_windows' is also set, the workstation
;		will not be closed at each return (you must obey AC_CLOSE).
;		This places a word == $FF00 at 'keep_wind_f', and if the top
;		byte is cleared the workstation closure is enabled again.
;
;NB:	Top bit of the word 'revise_type_f' is used for 3 purposes:
;   1:	Let ACC start 'exec_app' directly without awaiting event.
;   2:	Let APP delay 'exec_app' until an event occurs.
;   3:	Let APP loop back for more events after 'exec_app'
;The flag must be set by 'init_app' in the first two cases, and in case 3
;should be set/cleared in 'exec_app' to decide whether to exit program.
;
;Exit codes:	At exit from exec_app, d0 has the following effects
;		when the program was not started as accessory.
;
; negative	=> Pterm(d0)			=> error code exit
; 0		=> Pterm(d0)			=> error free exit
; 0x0000ADD0	=> Ptermres(tsr_size,0)		=> error free resident exit
; 0x0000ADD1	=> Ptermres(d1,0)		=> error free resident exit
; 0x0000ADD2	=> Ptermres(d1,d2)		=> error code resident exit
;
;----------------------------------------------------------------------------
;Start of:	init_app
;----------------------------------------------------------------------------
;
init_app:
;
	include		"uran\URAn_ARG.S"		;load ARGV handler
;
; Here URAn_ARG.S defines two variables
;
;arg_cnt	== number of strings on command line
;arg_ptr	-> NUL_separated argument strings + final NUL
;
;NB:	Both of the above include a dummy program name (first string==NUL).
;-------
	clr.l		VA_arg_p
	cmp		#2,arg_cnt
	blo.s		.done_arg
	move.l		arg_ptr,a0
	str_pass	a0
	move.l		a0,VA_arg_p
.done_arg:
;
;
;Add application dependent init above
;
	rts
;
;----------------------------------------------------------------------------
;End of:	init_app
;----------------------------------------------------------------------------
;Start of:	exec_app
;----------------------------------------------------------------------------
;
exec_app:
	tst.w		d7			;GEM/TOS program ?
	bmi.s		exec_TOS		;go run TOS/TTP code
exec_GEM:
	tst.l		d7			;APP/ACC program ?
	bmi		exec_multi_GEM		;go run ACC code
	cmp		#1,global+2		;Singletasking APP ?
	beq.s		exec_single_GEM		;then go show error message and exit program
exec_multi_APP:
	btst		#7,revise_type_f	;been here before ?
	bne		exec_multi_GEM		;then behave like an ACC this time
	bset		#7,revise_type_f	;tell main loop to behave like ACC
	bra		exit_exec_ok		;exit to main event loop
;
;NB: In multitasked APP mode, this code exits without reaching the user code
;    so as to let an autostarted APP simulate an ACC.  The user code will only
;    be reached on reception of VA_START or AC_OPEN messages.  This can easily
;    be changed however, by editing the code above.
;
;-------------------------------------
;
exec_TOS:
exit_exec_err:		;jump here to exit with error code -1
	moveq		#-1,d0
exit_terminate:		;jump here to exit with error code in d0
	bclr		#7,revise_type_f
	bra.s		exit_exec_app
;
;-------------------------------------
;
exec_single_GEM:
exec_multi_GEM:
exit_exec_ok:
	clr.l	d0
exit_exec_app:
	rts
;
;----------------------------------------------------------------------------
;End of:	exec_app
;----------------------------------------------------------------------------
;Start of:	exec_timer
;----------------------------------------------------------------------------
;
	.ifne		exec_timeout
exec_timer:
;
;
;Add application dependent timer event work above
;
exit_exec_timer:
	rts
	.endif		exec_timeout
;
;----------------------------------------------------------------------------
;End of:	exec_timer
;----------------------------------------------------------------------------
;Start of:	exec_mesag
;----------------------------------------------------------------------------
;
	.ifne		exec_message
exec_mesag:
	lea		message(pc),a3
	move		(a3),d0
	cmp		#AC_CLOSE,d0
	bne.s		.not_AC_CLOSE
	sf		keep_wind_f
	bra		exit_exec_mesag
;
.not_AC_CLOSE:
	cmp		#AP_TERM,d0
	bne.s		.not_AP_TERM
	tst.l		d7
	bmi		exit_exec_mesag		;AP_TERM is not for ACCs
	clr.l		d0			;flag no error
	bra		exit_terminate
;
.not_AP_TERM:
	cmp		#$4700,d0
	blo		exit_exec_mesag
	cmp		#$4780,d0
	bhs		exit_exec_mesag
	and		#$FF,d0
	lea		AVA_t(pc),a0
	tst.b		(a0,d0)
	bne.s		.fix_VA
.fix_AV:
	cmp.b		#AV_EXIT & $FF,d0
	bne.s		.not_AV_EXIT
	move		#-1,VA_partner_id
	bra.s		exit_exec_mesag
;
.not_AV_EXIT:
	move		2(a3),VA_partner_id
	lea		AV_message(pc),a4
	move		AV_partner_id(pc),d4
	bra		.fix_mess
;
.fix_VA:
	cmp		#VA_START & $FF,d0
	bne.s		.not_VA_START
	bsr		find_viewer
.not_VA_START:
	move		2(a3),AV_partner_id
	lea		VA_message(pc),a4
	move		VA_partner_id(pc),d4
	bmi.s		exit_exec_mesag
.fix_mess:
	move		(a3),(a4)
	move		global+4(pc),2(a4)
	clr		4(a4)
	move		6(a3),6(a4)
	move.l		8(a3),8(a4)
	move.l		12(a3),12(a4)
	gem_aes		appl_write,d4,#16,(a4)
exit_exec_mesag:
	rts
	.endif		exec_message
;
;----------------------------------------------------------------------------
;End of:	exec_mesag
;----------------------------------------------------------------------------
;Start of:	subroutines
;----------------------------------------------------------------------------
;
find_viewer:
	movem.l	d3-d4/a3-a4,-(sp)
	lea	viewer_t(pc),a4
	move.l	(a4)+,d3
.find_viewer_loop:
	move.l	(a4)+,a3		;a3 -> padded name of preferred application
	VA_set_partner	(a3)		;is the application accessible ?
	dbge	d3,.find_viewer_loop	;loop until found or preferences exhausted
	movem.l	(sp)+,d3-d4/a3-a4
	rts
;
;----------------------------------------------------------------------------
;End of:	subroutines
;----------------------------------------------------------------------------
;	Library function code will be expanded here
;
	make	SIM_links
	make	JAR_links
	make	AV_links
;
;----------------------------------------------------------------------------
;
text_limit:	.data
;
;----------------------------------------------------------------------------
;
viewer_t:
	dc.l	(viewer_t_end-viewer_t_ref)/4-1
viewer_t_ref:
	dc.l	pad8_EVER_s
	dc.l	pad8_1STG_s
	dc.l	pad8_1STV_s
	dc.l	pad8_STG_s
viewer_t_end:
;
pad8_EVER_s:
	dc.b	'EVEREST ',NUL
pad8_PCS_AV_s:
	dc.b	'PCS_AV  ',NUL
pad8_watch_it_s:
	dc.b	'WATCH_IT',NUL
pad8_STG_s:
	dc.b	'ST-GUIDE',NUL
pad8_1STG_s:
	dc.b	'1STGUIDE',NUL
pad8_1STV_s:
	dc.b	'1STVIEW ',NUL
;
STG_fname_s:
	dc.b	'C:\ST-GUIDE.ACC',NUL
my_name_s:
	dc.b	'PASS_AV ',NUL
	even
;
AVA_t:
	dc.b	0,1,0,0, 0,1,0,0, 0,0,0,0, 0,0,0,0
	dc.b	0,1,0,1, 0,1,0,1, 0,1,0,0, 0,0,0,0
	dc.b	0,1,0,1, 0,1,0,0, 0,1,0,0, 0,0,0,0
	dc.b	0,0,0,1, 0,1,0,0, 0,1,0,0, 0,0,0,0
;
	dc.b	0,1,0,0, 0,0,0,0, 0,0,0,0, 0,0,0,0
	dc.b	0,0,0,0, 1,0,1,0, 1,0,0,0, 0,0,0,0
	dc.b	1,0,0,0, 0,0,0,0, 0,0,0,0, 0,0,0,0
	dc.b	0,0,0,0, 0,0,0,0, 0,0,0,0, 0,0,0,0
;
;----------------------------------------------------------------------------
;
data_limit:	.bss
;
;----------------------------------------------------------------------------
;
VA_arg_p:	ds.l	1
VA_arg_s:	ds.b	MAX_VA_ARGLEN
		even
;
;
;Add uninitialized variables above
;
;----------------------------------------------------------------------------
bss_limit:	.end
;----------------------------------------------------------------------------
;End of file:	PASS_AVA.S
;----------------------------------------------------------------------------
